﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ButtonPusherEditor
{
    [Flags]
    enum Note : UInt32
    {
        None = 0b0000,
        SBT  = 0b0001,
        LBT  = 0b0010,
        SMO  = 0b0100,
        LMO  = 0b1000,

        ///////////////////////////////////////// SMO
        CP =       0b10000000000000000000000000000100,
        JP =       0b01000000000000000000000000000100,
        PU =       0b00100000000000000000000000000100,
        PI =       0b00010000000000000000000000000100,
        GD =       0b00001000000000000000000000000100,

        /////////////////////////////////////////L MO
        HU =       0b00000000000000000000000100001000,
        HD =       0b00000000000000000000000010001000,
        JS =       0b00000000000000000000000001001000,
        HP =       0b00000000000000000000000000101000,
        OT =       0b00000000000000000000000000011000,

        // Reset ////SSSSSSSSSSSSSSLLLLLLLLLLLLLLCCCC
        SMOReset = 0b00000000000000111111111111111011,
        LMOReset = 0b11111111111111000000000000000111
    }

    public partial class Form1 : Form
    {
        private static readonly int npm = 24;

        private float bpm = 120;

        private float measureScale = 2f;
        private readonly int MEASURE_DIVISOR = 8;

        private float MeasureSize
        {
            get
            {
                return measureScale * bpm;
            }
        }

        private List<Note[]> measures;

        private Pen btnPen;
        private Pen smoPen;
        private Pen lmoPen;
        private Pen measurePen;
        private Pen beatPen;

        public Form1()
        {
            InitializeComponent();
            DoubleBuffered = true;
            noteLane.Height *= 2;

            typeof(Panel).InvokeMember(
               "DoubleBuffered",
               BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.SetProperty,
               null,
               noteLane,
               new object[] { true });

            btnPen = new Pen(Color.FromArgb(0xA0, 0x00, 0x80, 0x00), 1f);
            smoPen = new Pen(Color.FromArgb(0xA0, 0x80, 0x00, 0x00), 1f);
            lmoPen = new Pen(Color.FromArgb(0xA0, 0x00, 0x00, 0xFF), 1f);

            measurePen = new Pen(Color.White, 3f);
            beatPen = new Pen(Color.White, 1f);

            ResetMeasures();
        }

        private void ResetMeasures()
        {
            measures = new List<Note[]>();
            for (int i = 0; i < 10; i++)
                AddMeasure();
        }

        private int? dragStartPointX = null;

        private void noteLane_MouseDown(object sender, MouseEventArgs e)
        {
            dragStartPointX = e.X;
        }

        private void noteLane_MouseUp(object sender, MouseEventArgs e)
        {
            if (dragStartPointX.HasValue)
            {
                bool isMotion = e.Button == MouseButtons.Right
                             || ModifierKeys.HasFlag(Keys.Control);

                SelectNote(isMotion, dragStartPointX.Value, e.X);
            }

            dragStartPointX = null;

            DrawMeasures();
        }

        private void SelectNote(bool isMotion, int posX, int posEndX = -1)
        {
            if (posEndX == -1)
                posEndX = posX;
            else if (posX > posEndX)
            {
                int tmp = posX;
                posX = posEndX;
                posEndX = tmp;
            }

            int measure = (int)(posX / MeasureSize);
            int beat = (int)Math.Round((posX - (MeasureSize * measure))
                                        * npm / MeasureSize);

            int measureEnd = measure;
            int beatEnd = beat;

            if (posEndX != -1)
            {
                measureEnd = (int)(posEndX / MeasureSize);
                beatEnd = (int)Math.Round((posEndX - (MeasureSize * measureEnd))
                                        * npm / MeasureSize);
            }

            while (measureEnd >= measures.Count)
                AddMeasure();

            if ((measures[measure][beat] & Note.SMO) != 0
                && isMotion && posX == posEndX)
            {
                measures[measure][beat] &= Note.SMOReset;
                return;
            }

            Note? motion = null;
            if (isMotion)
            {

                Func<string> MotionKeyDialog = delegate()
                {
                    Form prompt = new Form()
                    {
                        Width = 500,
                        Height = 150,
                        FormBorderStyle = FormBorderStyle.FixedDialog,
                        Text = "Input motion key",
                        StartPosition = FormStartPosition.CenterScreen
                    };
                    Label textLabel = new Label()
                    { Left = 50, Top = 20, Text = "Motion key" };
                    TextBox textBox = new TextBox()
                    { Left = 50, Top = 50, Width = 400 };
                    Button confirmation = new Button()
                    {
                        Text = "Ok",
                        Left = 350,
                        Width = 100,
                        Top = 70,
                        DialogResult = DialogResult.OK
                    };
                    confirmation.Click += (sender, e) => { prompt.Close(); };
                    prompt.Controls.Add(textBox);
                    prompt.Controls.Add(confirmation);
                    prompt.Controls.Add(textLabel);
                    prompt.AcceptButton = confirmation;

                    return prompt.ShowDialog() == DialogResult.OK
                                                ? textBox.Text : "";
                };

                string res = MotionKeyDialog();

                motion = Enum.GetValues(typeof(Note)).OfType<Note>()
                        .Where(x => x.ToString() == res.Trim())
                        .FirstOrDefault();

                if (motion == Note.None)
                {
                    MessageBox.Show("Motion not found: " + res);
                    return;
                }
            }

            if (posX == posEndX)
                FlipNoteFlag(measure, beat, motion);
            else
            {
                if (measure != measureEnd)
                {
                    for (int j = beat; j < npm; j++)
                        FlipNoteFlag(measure, j, motion, true);
                    for (int i = measure + 1; i < measureEnd; i++)
                    {
                        for (int j = 0; j < npm; j++)
                            FlipNoteFlag(i, j, motion, true);
                    }
                    for (int j = 0; j < beatEnd; j++)
                        FlipNoteFlag(measureEnd, j, motion, true);
                }
                else
                    for (int j = beat; j <= beatEnd; j++)
                        FlipNoteFlag(measure, j, motion, true);
            }
        }

        private void FlipNoteFlag(int measure, int beat, Note? motion = null, bool isLongMotion = false)
        {
            if (motion.HasValue)
            {
                if (isLongMotion)
                    measures[measure][beat] &= Note.LMOReset;
                else
                    measures[measure][beat] &= Note.SMOReset;
                
                measures[measure][beat] |= motion.Value;
            }
            else
                measures[measure][beat] ^= Note.SBT;
        }

        private void BPMChanged(object sender, EventArgs e)
        {
            if (float.TryParse(BPM.Text, out bpm))
            {
                DrawMeasures();
            }
        }
        
        private void AddMeasure(object sender = null, EventArgs e = null)
        {
            measures.Add(new Note[npm]);
            DrawMeasures();
        }

        private void DrawMeasures(object sender = null, PaintEventArgs e = null)
        {
            if (sender == null)
            {
                noteLane.Refresh();
                return;
            }

            var g = e.Graphics;

            for (int i = 0; i < measures.Count; i++)
            {
                g.DrawLine(measurePen, (i + 1) * MeasureSize, 0,
                                       (i + 1) * MeasureSize, noteLane.Height);

                for (int j = 0; j < npm; j++)
                {
                    float x = i * MeasureSize + j * MeasureSize / npm;

                    if (j % (npm / MEASURE_DIVISOR) == 0)
                        g.DrawLine(beatPen, x, 0, x, noteLane.Height);

                    if (measures[i][j] == Note.None)
                        continue;

                    bool btn = (measures[i][j] & (Note.SBT | Note.LBT)) != 0;
                    bool smo = (measures[i][j] & Note.SMO) != 0;
                    bool lmo = (measures[i][j] & Note.LMO) != 0;

                    if (btn)
                        g.DrawLine(btnPen, x, 0, x, noteLane.Height);
                    if (smo)
                        g.DrawLine(smoPen, x, 0, x, noteLane.Height);
                    if (lmo)
                        g.DrawLine(lmoPen, x, 0, x, noteLane.Height);
                }
            }
            
            noteLane.Width = (int)MeasureSize * measures.Count;
        }

        private void ExportBPE(object sender, EventArgs e)
        {
            var sfd = new SaveFileDialog
            {
                Filter = "Button Pusher Pattern|*.bpe"
            };

            if (sfd.ShowDialog() != DialogResult.OK || sfd.FileName == "")
                return;

            FileInfo file = new FileInfo(sfd.FileName);

            using (var writer
                = new StreamWriter((file.OpenWrite())))
            {
                writer.WriteLine("#TITLE " + Title.Text);
                writer.WriteLine("#ARTIST " + Artist.Text);
                writer.WriteLine("#GENRE " + Genre.Text);
                writer.WriteLine("#BPM " + BPM.Text);
                writer.WriteLine("#PLAYLEVEL " + Level.Text);
                writer.WriteLine();

                List<Note[]> tempMeasures = new List<Note[]>();
                foreach (Note[] measure in measures)
                {
                    Note[] tempMeasure = new Note[npm];
                    measure.CopyTo(tempMeasure, 0);
                    tempMeasures.Add(tempMeasure);
                }

                for (int i = 0; i < measures.Count; i++)
                {
                    Longbuttonize(i);
                    foreach (Note noteType in Enum.GetValues(typeof(Note)))
                    {
                        if (noteType > (Note)0b1111 || noteType == Note.None)
                            continue;

                        BPEDataLinePreprocess(noteType, i);
                    }
                }

                for (int i = 0; i < measures.Count; i++)
                {
                    foreach (Note noteType in Enum.GetValues(typeof(Note)))
                    {
                        if (noteType > (Note)0b1111 || noteType == Note.None)
                            continue;

                        string line = BPEDataLineBuilder(noteType, i);

                        writer.Write(line);
                    }
                }

                measures = tempMeasures;
            }
        }

        private void Longbuttonize(int line)
        {
            for (int i = 0; i < npm; i++)
            {
                int m, b;
                if (i != npm - 1)
                {
                    m = line;
                    b = i + 1;
                }
                else if (line < measures.Count - 1)
                {
                    m = line + 1;
                    b = 0;
                }
                else
                    continue;

                if ((measures[m][b] & (Note.SBT | Note.LBT)) != 0
                    && (measures[line][i] & (Note.SBT | Note.LBT)) != 0)
                {
                    measures[m][b] &= ~Note.SBT;
                    measures[m][b] |= Note.LBT;
                    measures[line][i] &= ~Note.SBT;
                    measures[line][i] |= Note.LBT;
                }
            }
        }

        bool keepLongBtn = false;
        bool keepLongMtn = false;

        private void BPEDataLinePreprocess(Note type, int line)
        {
            bool isLong = ((Note.LBT | Note.LMO) & type) != 0;
            bool isMotion = ((Note.SMO | Note.LMO) & type) != 0;

            if (!isLong)
                return;

            for (int i = 0; i < npm; i++)
            {
                int m, b;
                if (i != 0)
                {
                    m = line;
                    b = i - 1;
                }
                else if (line != 0)
                {
                    m = line - 1;
                    b = npm - 1;
                }
                else
                    continue;

                if ((measures[m][b] & measures[line][i] & type) != 0)
                {
                    if (!isMotion)
                    {
                        if (keepLongBtn)
                        {
                            measures[m][b] ^= type;
                        }
                        keepLongBtn = true;
                    }
                    else
                    {
                        if (keepLongMtn)
                        {
                            if ((measures[m][b] & measures[line][i]
                                & ~Note.LMOReset & ~(Note)0b1111) != 0)
                                measures[m][b] ^= measures[m][b] & ~Note.LMOReset;
                            else
                                keepLongMtn = false;
                        }
                        else
                            keepLongMtn = true;
                    }
                }
                else
                {
                    if (!isMotion)
                    {
                        if (keepLongBtn)
                        {
                            measures[line][i] ^=type;
                            measures[m][b] ^= type;
                        }
                        keepLongBtn = false;
                    }
                    else
                    {
                        if (keepLongMtn)
                        {
                            measures[line][i] ^= measures[m][b] & ~Note.LMOReset;
                            measures[m][b] ^= measures[m][b] & ~Note.LMOReset;
                        }
                        keepLongMtn = false;
                    }
                }
            }
        }

        private string BPEDataLineBuilder(Note type, int line)
        {
            int gcd = npm / 2;
            for (int i = 1; i < npm; i++)
            {
                if (gcd == 1)
                    break;
                if (i % gcd == 0)
                    continue;
                if ((measures[line][i] & type) != 0)
                {
                    i = 0;
                    gcd /= 2;
                    continue;
                }
            }

            string head = "#" + line.ToString().PadLeft(3, '0')
                + type.ToString() + " ";
            string body = "";

            for (int i = 0; i < npm / gcd; i++)
            {
                Note key = measures[line][i * gcd];
                if ((key & type) != 0)
                {
                    if ((type & (Note.SBT | Note.LBT)) != 0)
                        body += "AA";
                    else
                    {
                        foreach (Note nt in Enum.GetValues(typeof(Note)))
                        {
                            if (nt < (Note)0b1111
                                || nt == Note.LMOReset || nt==Note.SMOReset)
                                continue;
                            if ((key & nt & ~(Note)0b1111) != 0)
                                body += nt.ToString();
                        }
                    }
                }
                else
                    body += "00";
            }

            if (body == "0000")
                return "";

            return head + body + "\n";
        }

        private void ParseBPEHeader(FileInfo file)
        {
            using (StreamReader reader
                = new StreamReader(new BufferedStream(file.OpenRead())))
            {
                string line;
                while ((line = reader.ReadLine()) != null)
                {
                    if (line == "" || line[0] != '#')
                        continue;

                    string[] token = line.Split(new char[] { ' ' }, 2);
                    string field = token[0];
                    string value = token[1].Trim();

                    switch (field)
                    {
                        case "#TITLE":
                            Title.Text = value;
                            break;
                        case "#ARTIST":
                            Artist.Text = value;
                            break;
                        case "#GENRE":
                            Genre.Text = value;
                            break;
                        case "#BPM":
                            float outBPM;
                            if (float.TryParse(value, out outBPM))
                                bpm = outBPM;
                            else
                                bpm = 120;
                            BPM.Text = outBPM.ToString();
                            break;
                        case "#PLAYLEVEL":
                            int outLevel;
                            if (int.TryParse(value, out outLevel))
                                Level.Text = outLevel.ToString();
                            else
                                Level.Text = "0";
                            break;
                    }
                }
            }
        }

        private void ParseBPENote(FileInfo file)
        {
            ResetMeasures();

            Tuple<int, int> tempLongStart = null;

            using (StreamReader reader
                = new StreamReader(new BufferedStream(file.OpenRead())))
            {
                string line;
                while ((line = reader.ReadLine()) != null)
                {
                    if (line == "" || line[0] != '#')
                        continue;

                    string[] token = line.Split(new char[] { ' ' }, 2);
                    string field = token[0];
                    string value = token[1].Trim();

                    if (field.Length != 7)
                        continue;

                    string[] CHANNELS = { "SBT", "LBT", "SMO", "LMO" };

                    string channel = field.Substring(4, 3);

                    if (!int.TryParse(field.Substring(1, 3), out int measure))
                        continue;
                    if (Array.FindIndex(CHANNELS, x => x == channel) == -1)
                        continue;

                    int seq = value.Length / 2;

                    while (measure >= measures.Count)
                        AddMeasure();

                    for (int i = 0; i < seq; i++)
                    {
                        string key = value.Substring(i * 2, 2);
                        if (key == "00")
                            continue;

                        int beat = i * npm / seq;

                        if (channel == "SMO")
                        {
                            measures[measure][beat] |= Note.SMO;
                        }
                        else if (channel[0] == 'S')
                        {
                            measures[measure][beat] |= Note.SBT;
                        }
                        else if (channel[0] == 'L')
                        {
                            if (tempLongStart == null)
                            {
                                tempLongStart = new Tuple<int, int>(measure, beat);
                                continue;
                            }

                            if (measure == tempLongStart.Item1)
                            {
                                for (int b = tempLongStart.Item2; b < beat; b++)
                                    measures[measure][b] |= Note.SBT;
                            }
                            else
                            {
                                for (int b = tempLongStart.Item2; b < npm; b++)
                                    measures[tempLongStart.Item1][b] |= Note.SBT;
                                for (int m = tempLongStart.Item1 + 1; m < measure; m++)
                                {
                                    for (int b = 0; b < npm; b++)
                                        measures[m][b] |= Note.SBT;
                                }
                                for (int b = 0; b < beat; b++)
                                    measures[measure][b] |= Note.SBT;
                            }

                            tempLongStart = null;
                        }
                    }
                }
            }

            DrawMeasures();
        }

        private void ImportBPE(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog
            {
                Filter = "Button Pusher Pattern Files|*.bpe"
            };

            if (ofd.ShowDialog() != DialogResult.OK)
                return;

            FileInfo file = new FileInfo(ofd.FileName);

            ParseBPEHeader(file);
            ParseBPENote(file);
        }
    }
}
